#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_ILI9341.h"


#define TFT_CLK 13    // J2 pin 7        13
#define TFT_MISO 12    //    pin 9        12
#define TFT_MOSI 11    //    pin 6        11
#define TFT_CS   10    //    pin 3        10
#define TFT_DC    9    //    pin 5         9
#define TFT_RST   8    //    pin 4         8


#define LL 2
#define LM 10
#define RM 4
#define RR 5
#define REPLAY 6
#define START 7
#define BUZZER 3


#define LL_lane 0
#define LM_lane 1
#define RM_lane 2
#define RR_lane 3


#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define REST 0

// change this to make the song slower or faster
int tempo=144; 

int buzzer = 3;

int melody[] = {

  //Based on the arrangement at https://www.flutetunes.com/tunes.php?id=192
  //REST,4, REST,4, REST,4, REST,4,
  NOTE_E5, 4,  NOTE_B4,8,  NOTE_C5,8,  NOTE_D5,4,  NOTE_C5,8,  NOTE_B4,8,
  NOTE_A4, 4,  NOTE_A4,8,  NOTE_C5,8,  NOTE_E5,4,  NOTE_D5,8,  NOTE_C5,8,
  NOTE_B4, -4,  NOTE_C5,8,  NOTE_D5,4,  NOTE_E5,4,
  NOTE_C5, 4,  NOTE_A4,4,  NOTE_A4,8,  NOTE_A4,4,  NOTE_B4,8,  NOTE_C5,8,

  NOTE_D5, -4,  NOTE_F5,8,  NOTE_A5,4,  NOTE_G5,8,  NOTE_F5,8,
  NOTE_E5, -4,  NOTE_C5,8,  NOTE_E5,4,  NOTE_D5,8,  NOTE_C5,8,
  NOTE_B4, 4,  NOTE_B4,8,  NOTE_C5,8,  NOTE_D5,4,  NOTE_E5,4,
  NOTE_C5, 4,  NOTE_A4,4,  NOTE_A4,4, REST, 4//,

//  NOTE_E5, 4,  NOTE_B4,8,  NOTE_C5,8,  NOTE_D5,4,  NOTE_C5,8,  NOTE_B4,8,
//  NOTE_A4, 4,  NOTE_A4,8,  NOTE_C5,8,  NOTE_E5,4,  NOTE_D5,8,  NOTE_C5,8,
//  NOTE_B4, -4,  NOTE_C5,8,  NOTE_D5,4,  NOTE_E5,4,
//  NOTE_C5, 4,  NOTE_A4,4,  NOTE_A4,8,  NOTE_A4,4,  NOTE_B4,8,  NOTE_C5,8,
//
//  NOTE_D5, -4,  NOTE_F5,8,  NOTE_A5,4,  NOTE_G5,8,  NOTE_F5,8,
//  NOTE_E5, -4,  NOTE_C5,8,  NOTE_E5,4,  NOTE_D5,8,  NOTE_C5,8,
//  NOTE_B4, 4,  NOTE_B4,8,  NOTE_C5,8,  NOTE_D5,4,  NOTE_E5,4,
//  NOTE_C5, 4,  NOTE_A4,4,  NOTE_A4,4, REST, 4

  //REST,4, REST,4, REST,4, REST,4

};

int notes=sizeof(melody)/sizeof(melody[0])/2; 

// this calculates the duration of a whole note in ms (60s/tempo)*4 beats
int wholenote = (60000 * 4) / tempo;

int divider = 0, noteDuration = 0;

bool endOfSong = false;

void playSong(int thisNote) {
    // calculates the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) {
      // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      // dotted notes are represented with negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }

    // we only play the note for 90% of the duration, leaving 10% as a pause
    tone(buzzer, melody[thisNote], noteDuration*0.9);

    // Wait for the specief duration before playing the next note.
    delay(noteDuration);
    
    // stop the waveform generation before the next note.
    noTone(buzzer);
}

Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);

int width = tft.width();
int height = tft.height();


class NoteCircle 
{
  private:
    float x;
    float y;
    float prevX;
    float prevY;
  public:
    NoteCircle():
      x(-1),
      y(-1),
      prevX(-1),
      prevY(-1),
      lane(-1),
      onScreen(false),
      isBlank(false) 
      {}
    int rad = 7;
    const static int spd = 80;
    int getX() {return (int)this->x;}
    int getY() {return (int)this->y;}
    int getPY() {return (int)this->prevY;}
    int getPX() {return (int)this->prevX;}
    void setX(int x) {this->x = (float)x;}
    void setY(int y) {this->y = (float)y;}
    void setPY(int y) {this->prevY = (float)y;}
    void setPX(int x) {this->prevX = (float)x;}
    void updateY(int dt) {this->y += 10;}
    bool onScreen;
    bool isBlank;
    int lane;  
};


typedef struct {
  int x;
  int y;
  } notePos;

void drawPlayField();
void noteFalling(int lane);
void displayMenu();
void displayScore(int score);
void play(int dt);


notePos notePositions[] = {
  {width / 8, width / 8 + 20},
  {3 * width / 8, width / 8 + 20},
  {5 * width / 8, width / 8 + 20},
  {7 * width / 8, width / 8 + 20}
 };

int noteFallingLanes[] = {
  0, 0, 2, 2, 1, 1, 2, 2, 1, 1,
  3, 3, 2, 2, 0, 0, 1, 1, 1, 1
};

int lastTime = 0;

typedef enum {MENU, PLAY, SCORE} States;

States gameState = MENU;

bool blackScreen = false;
bool playStart = false;

int score = 0;

//displays current score
void displayScore(int score) {
  tft.fillRect(85, 0, 85, 18, ILI9341_BLACK);
  tft.setCursor(85, 0);
  tft.setTextSize(2);
  tft.print(score);
}


//draws the playing mode area
void drawPlayField() {
  if(!blackScreen) {
    tft.fillScreen(ILI9341_BLACK);
    blackScreen = true;
  }
  tft.drawLine(0, 20, width - 1, 20, ILI9341_WHITE);
  tft.drawLine(0, 20, 0, height, ILI9341_WHITE);
  //tft.drawLine(0, height, width - 1, height, ILI9341_WHITE);
  tft.drawLine(width - 1, height, width - 1, 20, ILI9341_WHITE);

  tft.drawChar(5, 0, 'S',ILI9341_WHITE, ILI9341_BLACK, 2);
  tft.drawChar(20, 0, 'C',ILI9341_WHITE, ILI9341_BLACK, 2);
  tft.drawChar(35, 0, 'O',ILI9341_WHITE, ILI9341_BLACK, 2);
  tft.drawChar(50, 0, 'R',ILI9341_WHITE, ILI9341_BLACK, 2);
  tft.drawChar(65, 0, 'E',ILI9341_WHITE, ILI9341_BLACK, 2);
  tft.drawChar(75, 0, ':',ILI9341_WHITE, ILI9341_BLACK, 2);

  displayScore(score);
  
  tft.drawLine(width / 4, 20, width / 4, height, ILI9341_WHITE);
  tft.drawLine(width / 2, 20, width / 2, height, ILI9341_WHITE);
  tft.drawLine(3 * width / 4, 20, 3 * width / 4, height, ILI9341_WHITE);

  tft.drawFastHLine(0, height - 50, width - 1, ILI9341_GREEN);
  tft.drawFastHLine(0, height - 5, width - 1, ILI9341_RED);
  
}

const int noCircles = 30;
NoteCircle circles[noCircles];
int laneIdx = 0;

typedef struct {
  bool pressed;
  int pin;
} btn;
btn buttons[4] = {
  {false, 2},
  {false, 10},
  {false, 4},
  {false, 5}
};
int lastBtnState[4];

int songIdx = 0;

//function to call in play mode
void play(int dt) {
  //find the first off-screen circle and put it on screen
  for (int i = 0; i < noCircles; i++) {
    if(!circles[i].onScreen) {
      circles[i].onScreen = true;
      int lane = noteFallingLanes[laneIdx];
      circles[i].setX(notePositions[lane].x);
      circles[i].setY(notePositions[lane].y);
      circles[i].lane = lane;

      laneIdx++;
      if (laneIdx > 19) {
        laneIdx = 0;
      }
      break;
    }
  }

  //check for any button presses
  for (int i = 0; i < 4; i++) {
    buttons[i].pressed = false;
    int crtState = digitalRead(buttons[i].pin);
    if (crtState != lastBtnState[i]) {
      if (crtState) {
        buttons[i].pressed = true;
      }
      lastBtnState[i] = crtState;
    }
  }

  //draw the notes on the screen
  for (int i = 0; i < noCircles; i++) {
    if (circles[i].onScreen) {
      if (!circles[i].isBlank) {
        //first flush the previous position of the note
        tft.fillCircle(circles[i].getPX(), circles[i].getPY(), circles[i].rad, ILI9341_BLACK);
        //draw it in the new position
        tft.fillCircle(circles[i].getX(), circles[i].getY(), circles[i].rad, ILI9341_WHITE);
      }
      //update the prev position
      circles[i].setPX(circles[i].getX());
      circles[i].setPY(circles[i].getY());
      //update the current position
      circles[i].updateY(dt);
      //mark off-screen notes as such
      if (circles[i].getY() > height) {
        circles[i].onScreen = false;
      }
      //check if button was pressed as the note is in the scoring area
      if (circles[i].getY() > height - 50 && circles[i].getY() < height - 5 && !circles[i].isBlank && buttons[circles[i].lane].pressed) {
        score += 5;
        playSong(songIdx);
        songIdx += 2;
        if (songIdx >= notes * 2) {
          endOfSong = true;
        }
        
      }
      //refresh the score
      displayScore(score);      
    }
  }
  //refresh the horizontal lines marking scoring area
  tft.drawFastHLine(0, height - 50, width - 1, ILI9341_GREEN);
  tft.drawFastHLine(0, height - 5, width - 1, ILI9341_RED);
}

//displays the start menu
void displayMenu() {
  if (!blackScreen) {
    tft.fillScreen(ILI9341_BLACK);
    blackScreen = true;
  }
  tft.setCursor(0, 0);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2.5);
  tft.println();
  tft.println();
  tft.println();
  tft.println("ARDUINO RHYTHM GAME");
  tft.setTextSize(1);
  tft.println("PRESS GREEN TO START");
  
}

//displays the final score
void displayFinalScore() {
  if (!blackScreen) {
    tft.fillScreen(ILI9341_BLACK);
    blackScreen = true;
  }
  tft.setCursor(0, 0);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2.5);
  tft.println();
  tft.println();
  tft.println();
  tft.println("ARDUINO RHYTHM GAME");
  tft.println("CONGRATULATIONS!");
  tft.println("YOUR SCORE:");
  tft.println(score);
  tft.setTextSize(1);
  tft.println("PRESS RED FOR REPLAY");
  tft.println("PRESS GREEN FOR START MENU");
  
}

int gameTime = 0;

void setup() {
  // put your setup code here, to run once:
  Serial.begin(9600);

  tft.begin();

  //set pins
  pinMode(LL, INPUT);
  pinMode(LM, INPUT);
  pinMode(RM, INPUT);
  pinMode(RR, INPUT);
  pinMode(REPLAY, INPUT);
  pinMode(START, INPUT);
  pinMode(BUZZER, OUTPUT);
  
  tft.fillScreen(ILI9341_BLACK);

  //set blanks on odd indexes
  for (int i = 0; i < noCircles; i++) {
    if (i%2 == 1) {
      circles[i].isBlank = true;
    }
  }

  
}

void loop() {
  // put your main code here, to run repeatedly:

  int currentTime = millis();
  int dt = currentTime - lastTime;
  lastTime = currentTime;

  gameTime += dt; 

  if (gameTime > 50000) {
    endOfSong = true;
  }
  
  if (gameState == MENU) {
    displayMenu();
    if (digitalRead(START)) {
      gameState = PLAY;
      blackScreen = false;
      playStart = false;
      gameTime = 0;
    }
  }

  if(gameState == PLAY) {
    if (!playStart) {
      drawPlayField();
      playStart = true;
    }
    play(dt);
    if(endOfSong) {
      gameState = SCORE;
      blackScreen = false;
    }
    
  }

  if(gameState == SCORE) {
    displayFinalScore();
    if (digitalRead(START)) {
      gameState = MENU;
      blackScreen = false;
      score = 0;
      songIdx = 0;
    }
    if(digitalRead(REPLAY)) {
      gameState = PLAY;
      blackScreen = false;
      playStart = false;
      score = 0;
      songIdx = 0;
    }
  }
}
  
